"""Test code for categories."""

import pytest

from icalendar.prop import vCategory, vText


# we also work on text
@pytest.fixture(params=[vCategory, vText])
def v_cat(request):
    """Fixture for vCategory and vText."""
    return request.param


def test_categories_parse_with_no_items(v_cat):
    """We make sure all items are collected."""
    cats = v_cat.from_jcal(["categories", {}, "text", ""])
    assert cats.cats == [""]


def test_categories_parse_with_more_items(v_cat):
    """We make sure all items are collected."""
    cats = v_cat.from_jcal(["categories", {}, "text", "WORK", "PERSONAL", "URGENT"])
    assert cats.cats == ["WORK", "PERSONAL", "URGENT"]


def test_category_jcal_with_1_item():
    """Check converting vCategory to and from jCal."""
    cats = vCategory(["item 1"])
    jcal = cats.to_jcal("CATEGORIES")
    assert jcal == ["CATEGORIES", {}, "text", "item 1"]


def test_category_jcal_with_0_items():
    """Check converting vCategory to and from jCal."""
    cats = vCategory([])
    jcal = cats.to_jcal("CATEGORIES")
    assert jcal == ["CATEGORIES", {}, "text", ""]


def test_category_jcal_with_1_empty_item():
    """Check converting vCategory to and from jCal."""
    cats = vCategory([""])
    jcal = cats.to_jcal("CATEGORIES")
    assert jcal == ["CATEGORIES", {}, "text", ""]


def test_category_jcal_with_3_items():
    """Check converting vCategory to and from jCal."""
    cats = vCategory(["item 1", "item 2", "item 3"])
    jcal = cats.to_jcal("CATEGORIES")
    assert jcal == ["CATEGORIES", {}, "text", "item 1", "item 2", "item 3"]



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

from icalendar.tools import to_datetime

from .windows_to_olson import WINDOWS_TO_OLSON

if TYPE_CHECKING:
    import datetime

    from dateutil.rrule import rrule

    from icalendar import prop
    from icalendar.cal import Timezone

    from .provider import TZProvider

DEFAULT_TIMEZONE_PROVIDER = "zoneinfo"


class TZP:
    """This is the timezone provider proxy.

    If you would like to have another timezone implementation,
    you can create a new one and pass it to this proxy.
    All of icalendar will then use this timezone implementation.
    """

    def __init__(self, provider: Union[str, TZProvider] = DEFAULT_TIMEZONE_PROVIDER):
        """Create a new timezone implementation proxy."""
        self.use(provider)

    def use_pytz(self) -> None:
        """Use pytz as the timezone provider."""
        from .pytz import PYTZ

        self._use(PYTZ())

    def use_zoneinfo(self) -> None:
        """Use zoneinfo as the timezone provider."""
        from .zoneinfo import ZONEINFO

        self._use(ZONEINFO())

    def _use(self, provider: TZProvider) -> None:
        """Use a timezone implementation."""
        self.__tz_cache = {}
        self.__provider = provider

    def use(self, provider: Union[str, TZProvider]):
        """Switch to a different timezone provider."""
        if isinstance(provider, str):
            use_provider = getattr(self, f"use_{provider}", None)
            if use_provider is None:
                raise ValueError(
                    f"Unknown provider {provider}. Use 'pytz' or 'zoneinfo'."
                )
            use_provider()
        else:
            self._use(provider)

    def use_default(self):
        """Use the default timezone provider."""
        self.use(DEFAULT_TIMEZONE_PROVIDER)

    def localize_utc(self, dt: datetime.date) -> datetime.datetime:
        """Return the datetime in UTC.

        If the datetime has no timezone, set UTC as its timezone.
        """
        return self.__provider.localize_utc(to_datetime(dt))

    def localize(
        self, dt: datetime.date, tz: Union[datetime.tzinfo, str, None]
    ) -> datetime.datetime:
        """Localize a datetime to a timezone."""
        if isinstance(tz, str):
            tz = self.timezone(tz)
        if tz is None:
            return dt.replace(tzinfo=None)
        return self.__provider.localize(to_datetime(dt), tz)

    def cache_timezone_component(self, timezone_component: Timezone.Timezone) -> None:
        """Cache the timezone that is created from a timezone component
        if it is not already known.

        This can influence the result from timezone(): Once cached, the
        custom timezone is returned from timezone().
        """
        _unclean_id = timezone_component["TZID"]
        _id = self.clean_timezone_id(_unclean_id)
        if (
            not self.__provider.knows_timezone_id(_id)
            and not self.__provider.knows_timezone_id(_unclean_id)
            and _id not in self.__tz_cache
        ):
            self.__tz_cache[_id] = timezone_component.to_tz(self, lookup_tzid=False)

    def fix_rrule_until(self, rrule: rrule, ical_rrule: prop.vRecur) -> None:
        """Make sure the until value works."""
        self.__provider.fix_rrule_until(rrule, ical_rrule)

    def create_timezone(self, timezone_component: Timezone.Timezone) -> datetime.tzinfo:
        """Create a timezone from a timezone component.

        This component will not be cached.
        """
        return self.__provider.create_timezone(timezone_component)

    def clean_timezone_id(self, tzid: str) -> str:
        """Return a clean version of the timezone id.

        Timezone ids can be a bit unclean, starting with a / for example.
        Internally, we should use this to identify timezones.
        """
        return tzid.strip("/")

    def timezone(self, tz_id: str) -> Optional[datetime.tzinfo]:
        """Return a timezone with an id or None if we cannot find it."""
        _unclean_id = tz_id
        tz_id = self.clean_timezone_id(tz_id)
        tz = self.__provider.timezone(tz_id)
        if tz is not None:
            return tz
        if tz_id in WINDOWS_TO_OLSON:
            tz = self.__provider.timezone(WINDOWS_TO_OLSON[tz_id])
        return tz or self.__provider.timezone(_unclean_id) or self.__tz_cache.get(tz_id)

    def uses_pytz(self) -> bool:
        """Whether we use pytz at all."""
        return self.__provider.uses_pytz()

    def uses_zoneinfo(self) -> bool:
        """Whether we use zoneinfo."""
        return self.__provider.uses_zoneinfo()

    @property
    def name(self) -> str:
        """The name of the timezone component used."""
        return self.__provider.name

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name!r})"


__all__ = ["TZP"]



""":rfc:`5545` iCalendar component."""

from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING, Sequence

from icalendar.attr import (
    CONCEPTS_TYPE_SETTER,
    LINKS_TYPE_SETTER,
    RELATED_TO_TYPE_SETTER,
    categories_property,
    images_property,
    multi_language_text_property,
    single_string_property,
    source_property,
    uid_property,
    url_property,
)
from icalendar.cal.component import Component
from icalendar.cal.examples import get_example
from icalendar.cal.timezone import Timezone
from icalendar.error import IncompleteComponent
from icalendar.version import __version__

if TYPE_CHECKING:
    import uuid
    from datetime import date, datetime

    from icalendar.cal.availability import Availability
    from icalendar.cal.event import Event
    from icalendar.cal.free_busy import FreeBusy
    from icalendar.cal.todo import Todo


class Calendar(Component):
    """
        The "VCALENDAR" object is a collection of calendar information.
        This information can include a variety of components, such as
        "VEVENT", "VTODO", "VJOURNAL", "VFREEBUSY", "VTIMEZONE", or any
        other type of calendar component.

    Examples:
        Create a new Calendar:

            >>> from icalendar import Calendar
            >>> calendar = Calendar.new(name="My Calendar")
            >>> print(calendar.calendar_name)
            My Calendar

    """

    name = "VCALENDAR"
    canonical_order = (
        "VERSION",
        "PRODID",
        "CALSCALE",
        "METHOD",
        "DESCRIPTION",
        "X-WR-CALDESC",
        "NAME",
        "X-WR-CALNAME",
    )
    required = (
        "PRODID",
        "VERSION",
    )
    singletons = (
        "PRODID",
        "VERSION",
        "CALSCALE",
        "METHOD",
        "COLOR",  # RFC 7986
    )
    multiple = (
        "CATEGORIES",  # RFC 7986
        "DESCRIPTION",  # RFC 7986
        "NAME",  # RFC 7986
    )

    @classmethod
    def example(cls, name: str = "example") -> Calendar:
        """Return the calendar example with the given name."""
        return cls.from_ical(get_example("calendars", name))

    @classmethod
    def from_ical(cls, st, multiple=False):
        comps = Component.from_ical(st, multiple=True)
        all_timezones_so_far = True
        for comp in comps:
            for component in comp.subcomponents:
                if component.name == "VTIMEZONE":
                    if not all_timezones_so_far:
                        # If a preceding component refers to a VTIMEZONE defined
                        # later in the source st
                        # (forward references are allowed by RFC 5545), then the
                        # earlier component may have
                        # the wrong timezone attached.
                        # However, during computation of comps, all VTIMEZONEs
                        # observed do end up in
                        # the timezone cache. So simply re-running from_ical will
                        # rely on the cache
                        # for those forward references to produce the correct result.
                        # See test_create_america_new_york_forward_reference.
                        return Component.from_ical(st, multiple)
                else:
                    all_timezones_so_far = False

        # No potentially forward VTIMEZONEs to worry about
        if multiple:
            return comps
        if len(comps) > 1:
            raise ValueError(
                cls._format_error(
                    "Found multiple components where only one is allowed", st
                )
            )
        if len(comps) < 1:
            raise ValueError(
                cls._format_error(
                    "Found no components where exactly one is required", st
                )
            )
        return comps[0]

    @property
    def events(self) -> list[Event]:
        """All event components in the calendar.

        This is a shortcut to get all events.
        Modifications do not change the calendar.
        Use :py:meth:`Component.add_component`.

        >>> from icalendar import Calendar
        >>> calendar = Calendar.example()
        >>> event = calendar.events[0]
        >>> event.start
        datetime.date(2022, 1, 1)
        >>> print(event["SUMMARY"])
        New Year's Day
        """
        return self.walk("VEVENT")

    @property
    def todos(self) -> list[Todo]:
        """All todo components in the calendar.

        This is a shortcut to get all todos.
        Modifications do not change the calendar.
        Use :py:meth:`Component.add_component`.
        """
        return self.walk("VTODO")

    @property
    def availabilities(self) -> list[Availability]:
        """All :class:`Availability` components in the calendar.

        This is a shortcut to get all availabilities.
        Modifications do not change the calendar.
        Use :py:meth:`Component.add_component`.
        """
        return self.walk("VAVAILABILITY")

    @property
    def freebusy(self) -> list[FreeBusy]:
        """All FreeBusy components in the calendar.

        This is a shortcut to get all FreeBusy.
        Modifications do not change the calendar.
        Use :py:meth:`Component.add_component`.
        """
        return self.walk("VFREEBUSY")

    def get_used_tzids(self) -> set[str]:
        """The set of TZIDs in use.

        This goes through the whole calendar to find all occurrences of
        timezone information like the TZID parameter in all attributes.

        >>> from icalendar import Calendar
        >>> calendar = Calendar.example("timezone_rdate")
        >>> calendar.get_used_tzids()
        {'posix/Europe/Vaduz'}

        Even if you use UTC, this will not show up.
        """
        result = set()
        for _name, value in self.property_items(sorted=False):
            if hasattr(value, "params"):
                result.add(value.params.get("TZID"))
        return result - {None}

    def get_missing_tzids(self) -> set[str]:
        """The set of missing timezone component tzids.

        To create a :rfc:`5545` compatible calendar,
        all of these timezones should be added.
        """
        tzids = self.get_used_tzids()
        for timezone in self.timezones:
            tzids.remove(timezone.tz_name)
        return tzids

    @property
    def timezones(self) -> list[Timezone]:
        """Return the timezones components in this calendar.

        >>> from icalendar import Calendar
        >>> calendar = Calendar.example("pacific_fiji")
        >>> [timezone.tz_name for timezone in calendar.timezones]
        ['custom_Pacific/Fiji']

        .. note::

            This is a read-only property.
        """
        return self.walk("VTIMEZONE")

    def add_missing_timezones(
        self,
        first_date: date = Timezone.DEFAULT_FIRST_DATE,
        last_date: date = Timezone.DEFAULT_LAST_DATE,
    ):
        """Add all missing VTIMEZONE components.

        This adds all the timezone components that are required.
        VTIMEZONE components are inserted at the beginning of the calendar
        to ensure they appear before other components that reference them.

        .. note::

            Timezones that are not known will not be added.

        :param first_date: earlier than anything that happens in the calendar
        :param last_date: later than anything happening in the calendar

        >>> from icalendar import Calendar, Event
        >>> from datetime import datetime
        >>> from zoneinfo import ZoneInfo
        >>> calendar = Calendar()
        >>> event = Event()
        >>> calendar.add_component(event)
        >>> event.start = datetime(1990, 10, 11, 12, tzinfo=ZoneInfo("Europe/Berlin"))
        >>> calendar.timezones
        []
        >>> calendar.add_missing_timezones()
        >>> calendar.timezones[0].tz_name
        'Europe/Berlin'
        >>> calendar.get_missing_tzids()  # check that all are added
        set()
        """
        missing_tzids = self.get_missing_tzids()
        if not missing_tzids:
            return

        existing_timezone_count = len(self.timezones)

        for tzid in missing_tzids:
            try:
                timezone = Timezone.from_tzid(
                    tzid, first_date=first_date, last_date=last_date
                )
            except ValueError:
                continue
            self.subcomponents.insert(existing_timezone_count, timezone)
            existing_timezone_count += 1

    calendar_name = multi_language_text_property(
        "NAME",
        "X-WR-CALNAME",
        """This property specifies the name of the calendar.

    This implements :rfc:`7986` ``NAME`` and ``X-WR-CALNAME``.

    Property Parameters:
        IANA, non-standard, alternate text
        representation, and language property parameters can be specified
        on this property.

    Conformance:
        This property can be specified multiple times in an
        iCalendar object.  However, each property MUST represent the name
        of the calendar in a different language.

    Description:
        This property is used to specify a name of the
        iCalendar object that can be used by calendar user agents when
        presenting the calendar data to a user.  Whilst a calendar only
        has a single name, multiple language variants can be specified by
        including this property multiple times with different "LANGUAGE"
        parameter values on each.

    Example:
        Below, we set the name of the calendar.

        .. code-block:: pycon

            >>> from icalendar import Calendar
            >>> calendar = Calendar()
            >>> calendar.calendar_name = "My Calendar"
            >>> print(calendar.to_ical())
            BEGIN:VCALENDAR
            NAME:My Calendar
            END:VCALENDAR
    """,
    )

    description = multi_language_text_property(
        "DESCRIPTION",
        "X-WR-CALDESC",
        """This property specifies the description of the calendar.

    This implements :rfc:`7986` ``DESCRIPTION`` and ``X-WR-CALDESC``.

    Conformance:
        This property can be specified multiple times in an
        iCalendar object.  However, each property MUST represent the
        description of the calendar in a different language.

    Description:
        This property is used to specify a lengthy textual
        description of the iCalendar object that can be used by calendar
        user agents when describing the nature of the calendar data to a
        user.  Whilst a calendar only has a single description, multiple
        language variants can be specified by including this property
        multiple times with different "LANGUAGE" parameter values on each.

    Example:
        Below, we add a description to a calendar.

        .. code-block:: pycon

            >>> from icalendar import Calendar
            >>> calendar = Calendar()
            >>> calendar.description = "This is a calendar"
            >>> print(calendar.to_ical())
            BEGIN:VCALENDAR
            DESCRIPTION:This is a calendar
            END:VCALENDAR
    """,
    )

    color = single_string_property(
        "COLOR",
        """This property specifies a color used for displaying the calendar.

    This implements :rfc:`7986` ``COLOR`` and ``X-APPLE-CALENDAR-COLOR``.
    Please note that since :rfc:`7986`, subcomponents can have their own color.

    Property Parameters:
        IANA and non-standard property parameters can
        be specified on this property.

    Conformance:
        This property can be specified once in an iCalendar
        object or in ``VEVENT``, ``VTODO``, or ``VJOURNAL`` calendar components.

    Description:
        This property specifies a color that clients MAY use
        when presenting the relevant data to a user.  Typically, this
        would appear as the "background" color of events or tasks.  The
        value is a case-insensitive color name taken from the CSS3 set of
        names, defined in Section 4.3 of `W3C.REC-css3-color-20110607 <https://www.w3.org/TR/css-color-3/>`_.

    Example:
        ``"turquoise"``, ``"#ffffff"``

        .. code-block:: pycon

            >>> from icalendar import Calendar
            >>> calendar = Calendar()
            >>> calendar.color = "black"
            >>> print(calendar.to_ical())
            BEGIN:VCALENDAR
            COLOR:black
            END:VCALENDAR

    """,
        "X-APPLE-CALENDAR-COLOR",
    )
    categories = categories_property
    uid = uid_property
    prodid = single_string_property(
        "PRODID",
        """PRODID specifies the identifier for the product that created the iCalendar object.

Conformance:
    The property MUST be specified once in an iCalendar object.

Description:
    The vendor of the implementation SHOULD assure that
    this is a globally unique identifier; using some technique such as
    an FPI value, as defined in [ISO.9070.1991].

    This property SHOULD NOT be used to alter the interpretation of an
    iCalendar object beyond the semantics specified in this memo.  For
    example, it is not to be used to further the understanding of non-
    standard properties.

Example:
    The following is an example of this property. It does not
    imply that English is the default language.

    .. code-block:: text

        -//ABC Corporation//NONSGML My Product//EN
""",  # noqa: E501
    )
    version = single_string_property(
        "VERSION",
        """VERSION of the calendar specification.

The default is ``"2.0"`` for :rfc:`5545`.

Purpose:
    This property specifies the identifier corresponding to the
    highest version number or the minimum and maximum range of the
    iCalendar specification that is required in order to interpret the
    iCalendar object.


      """,
    )

    calscale = single_string_property(
        "CALSCALE",
        """CALSCALE defines the calendar scale used for the calendar information specified in the iCalendar object.

Compatibility:
    :rfc:`7529` makes the case that GREGORIAN stays the default and other calendar scales
    are implemented on the RRULE.

Conformance:
    This property can be specified once in an iCalendar
    object.  The default value is "GREGORIAN".

Description:
    This memo is based on the Gregorian calendar scale.
    The Gregorian calendar scale is assumed if this property is not
    specified in the iCalendar object.  It is expected that other
    calendar scales will be defined in other specifications or by
    future versions of this memo.
        """,  # noqa: E501
        default="GREGORIAN",
    )
    method = single_string_property(
        "METHOD",
        """METHOD defines the iCalendar object method associated with the calendar object.

Description:
    When used in a MIME message entity, the value of this
    property MUST be the same as the Content-Type "method" parameter
    value.  If either the "METHOD" property or the Content-Type
    "method" parameter is specified, then the other MUST also be
    specified.

    No methods are defined by this specification.  This is the subject
    of other specifications, such as the iCalendar Transport-
    independent Interoperability Protocol (iTIP) defined by :rfc:`5546`.

    If this property is not present in the iCalendar object, then a
    scheduling transaction MUST NOT be assumed.  In such cases, the
    iCalendar object is merely being used to transport a snapshot of
    some calendar information; without the intention of conveying a
    scheduling semantic.
""",  # noqa: E501
    )
    url = url_property
    source = source_property

    @property
    def refresh_interval(self) -> timedelta | None:
        """REFRESH-INTERVAL specifies a suggested minimum interval for
        polling for changes of the calendar data from the original source
        of that data.

        Conformance:
            This property can be specified once in an iCalendar
            object, consisting of a positive duration of time.

        Description:
            This property specifies a positive duration that gives
            a suggested minimum polling interval for checking for updates to
            the calendar data.  The value of this property SHOULD be used by
            calendar user agents to limit the polling interval for calendar
            data updates to the minimum interval specified.

        Raises:
            ValueError: When setting a negative duration.
        """
        refresh_interval = self.get("REFRESH-INTERVAL")
        return refresh_interval.dt if refresh_interval else None

    @refresh_interval.setter
    def refresh_interval(self, value: timedelta | None):
        """Set the REFRESH-INTERVAL."""
        if not isinstance(value, timedelta) and value is not None:
            raise TypeError(
                "REFRESH-INTERVAL must be either a positive timedelta,"
                " or None to delete it."
            )
        if value is not None and value.total_seconds() <= 0:
            raise ValueError("REFRESH-INTERVAL must be a positive timedelta.")
        if value is not None:
            del self.refresh_interval
            self.add("REFRESH-INTERVAL", value)
        else:
            del self.refresh_interval

    @refresh_interval.deleter
    def refresh_interval(self):
        """Delete REFRESH-INTERVAL."""
        self.pop("REFRESH-INTERVAL")

    images = images_property

    @classmethod
    def new(
        cls,
        /,
        calscale: str | None = None,
        categories: Sequence[str] = (),
        color: str | None = None,
        concepts: CONCEPTS_TYPE_SETTER = None,
        description: str | None = None,
        language: str | None = None,
        last_modified: date | datetime | None = None,
        links: LINKS_TYPE_SETTER = None,
        method: str | None = None,
        name: str | None = None,
        organization: str | None = None,
        prodid: str | None = None,
        refresh_interval: timedelta | None = None,
        refids: list[str] | str | None = None,
        related_to: RELATED_TO_TYPE_SETTER = None,
        source: str | None = None,
        uid: str | uuid.UUID | None = None,
        url: str | None = None,
        version: str = "2.0",
    ):
        """Create a new Calendar with all required properties.

        This creates a new Calendar in accordance with :rfc:`5545` and :rfc:`7986`.

        Arguments:
            calscale: The :attr:`calscale` of the calendar.
            categories: The :attr:`categories` of the calendar.
            color: The :attr:`color` of the calendar.
            concepts: The :attr:`~icalendar.Component.concepts` of the calendar.
            description: The :attr:`description` of the calendar.
            language: The language for the calendar. Used to generate localized `prodid`.
            last_modified: The :attr:`~icalendar.Component.last_modified` of the calendar.
            links: The :attr:`~icalendar.Component.links` of the calendar.
            method: The :attr:`method` of the calendar.
            name: The :attr:`calendar_name` of the calendar.
            organization: The organization name. Used to generate `prodid` if not provided.
            prodid: The :attr:`prodid` of the component. If None and organization is provided,
                generates a `prodid` in format "-//organization//name//language".
            refresh_interval: The :attr:`refresh_interval` of the calendar.
            refids: :attr:`~icalendar.Component.refids` of the calendar.
            related_to: :attr:`~icalendar.Component.related_to` of the calendar.
            source: The :attr:`source` of the calendar.
            uid: The :attr:`uid` of the calendar.
                If None, this is set to a new :func:`uuid.uuid4`.
            url: The :attr:`url` of the calendar.
            version: The :attr:`version` of the calendar.

        Returns:
            :class:`Calendar`

        Raises:
            InvalidCalendar: If the content is not valid according to :rfc:`5545`.

        .. warning:: As time progresses, we will be stricter with the validation.
        """  # noqa: E501
        calendar: Calendar = super().new(
            last_modified=last_modified,
            links=links,
            related_to=related_to,
            refids=refids,
            concepts=concepts,
        )

        # Generate prodid if not provided but organization is given
        if prodid is None and organization:
            app_name = name or "Calendar"
            lang = language.upper() if language else "EN"
            prodid = f"-//{organization}//{app_name}//{lang}"
        elif prodid is None:
            prodid = f"-//collective//icalendar//{__version__}//EN"

        calendar.prodid = prodid
        calendar.version = version
        calendar.calendar_name = name
        calendar.color = color
        calendar.description = description
        calendar.method = method
        calendar.calscale = calscale
        calendar.categories = categories
        calendar.uid = uid
        calendar.url = url
        calendar.refresh_interval = refresh_interval
        calendar.source = source

        return calendar

    def validate(self):
        """Validate that the calendar has required properties and components.

        This method can be called explicitly to validate a calendar before output.

        Raises:
            IncompleteComponent: If the calendar lacks required properties or
                components.
        """
        if not self.get("PRODID"):
            raise IncompleteComponent("Calendar must have a PRODID")
        if not self.get("VERSION"):
            raise IncompleteComponent("Calendar must have a VERSION")
        if not self.subcomponents:
            raise IncompleteComponent(
                "Calendar must contain at least one component (event, todo, etc.)"
            )


__all__ = ["Calendar"]



"""Test getting the rrules from a component."""

import pytest

from icalendar import (
    vRecur,
)
from icalendar.cal.component import Component
from icalendar.cal.event import Event
from icalendar.cal.journal import Journal
from icalendar.cal.timezone import TimezoneDaylight, TimezoneStandard
from icalendar.cal.todo import Todo

RRULE_0 = vRecur.from_ical("FREQ=DAILY;COUNT=10")
RRULE_1 = vRecur.from_ical("FREQ=DAILY;UNTIL=19971224T000000Z")
RRULE_2 = vRecur.from_ical("FREQ=DAILY;INTERVAL=2")
RRULE_3 = vRecur.from_ical("FREQ=DAILY;INTERVAL=10;COUNT=5")
RRULE_4 = vRecur.from_ical(
    "FREQ=YEARLY;UNTIL=20000131T140000Z;BYMONTH=1;BYDAY=SU,MO,TU,WE,TH,FR,SA"
)


@pytest.fixture(params=[RRULE_0, RRULE_1, RRULE_2, RRULE_3, RRULE_4])
def rrule(request) -> str:
    """An rrule."""
    return request.param


@pytest.fixture(params=[Event, Todo, Journal, TimezoneDaylight, TimezoneStandard])
def c_rrule(request) -> Component:
    """Return a component that uses RDATE."""
    return request.param()


def test_no_rrules_by_default(c_rrule):
    """We expect no rdate by default."""
    assert c_rrule.rrules == []


def test_one_rrule(c_rrule, rrule):
    """Add one rrule."""
    c_rrule.add("rrule", rrule)
    assert c_rrule.rrules == [rrule]


def test_two_rrules(c_rrule, rrule):
    """Add two rrules."""
    c_rrule.add("rrule", rrule)
    c_rrule.add("rrule", RRULE_3)
    assert c_rrule.rrules == [rrule, RRULE_3]



"""Issue #27 - multiple periods

https://github.com/collective/icalendar/issues/27
"""


def test_issue_27_multiple_periods(calendars):
    free_busy = list(
        calendars.issue_27_multiple_periods_in_freebusy_multiple_freebusies.walk(
            "VFREEBUSY"
        )
    )[0]
    free_busy_period = free_busy["freebusy"]
    print(free_busy["freebusy"])
    equivalent_way_of_defining_free_busy = list(
        calendars.issue_27_multiple_periods_in_freebusy_one_freebusy.walk("VFREEBUSY")
    )[0]
    free_busy_period_equivalent = equivalent_way_of_defining_free_busy["freebusy"]
    assert free_busy_period == free_busy_period_equivalent



"""Test the VALARM compatibility of RFC 9074.

See https://www.rfc-editor.org/rfc/rfc9074.html
and also https://github.com/collective/icalendar/issues/657
"""

import pytest

from icalendar.prop import vDDDTypes, vText


@pytest.mark.parametrize(
    ("prop", "cls", "file", "alarm_index"),
    [
        ("UID", vText, "rfc_9074_example_1", 0),
        ("RELATED-TO", vText, "rfc_9074_example_2", 1),
        ("ACKNOWLEDGED", vDDDTypes, "rfc_9074_example_3", 0),
        ("PROXIMITY", vText, "rfc_9074_example_proximity", 0),
    ],
)
def test_calendar_types(events, prop, cls, file, alarm_index):
    """Check the types of the properties."""
    event = events[file]
    alarm = event.subcomponents[alarm_index]
    value = alarm[prop]
    assert isinstance(value, cls)


def test_snooze(events):
    """Check values of the alarms."""
    alarm_1 = events.rfc_9074_example_3.subcomponents[0]
    assert alarm_1["ACKNOWLEDGED"].dt == vDDDTypes.from_ical("20210302T152024Z")
    alarm_2 = events.rfc_9074_example_3.subcomponents[1]
    assert alarm_2["RELATED-TO"] == "8297C37D-BA2D-4476-91AE-C1EAA364F8E1"
    assert alarm_2["RELATED-TO"].params["RELTYPE"] == "SNOOZE"


def test_proximity(events):
    """Check the proximity values."""
    alarm = events.rfc_9074_example_proximity.subcomponents[0]
    assert alarm["PROXIMITY"] == "DEPART"
    assert len(alarm.subcomponents) == 1
    location = alarm.subcomponents[0]
    assert location["UID"] == "123456-abcdef-98765432"



"""Test for Issue #321 - AssertionError with VTIMEZONE containing only DAYLIGHT.

When a VTIMEZONE component contains only a DAYLIGHT subcomponent (no STANDARD),
converting it to a pytz timezone with lookup_tzid=False would fail with an
AssertionError.

The issue occurs because the code assumes it can find a non-DST (STANDARD)
transition to calculate the dst_offset, but when only DAYLIGHT exists, this
assumption fails.

Original issue: https://github.com/collective/icalendar/issues/321
Original PR with test case: https://github.com/collective/icalendar/pull/379
Test case by: @niccokunzmann
Fix by: @SashankBhamidi
"""

import pytest


class TestIssue321DstOffset:
    """Test VTIMEZONE conversion with only DAYLIGHT component."""

    @pytest.fixture
    def calendar_with_only_daylight(self, calendars):
        """Load calendar with VTIMEZONE containing only DAYLIGHT."""
        return calendars.issue_321_assert_dst_offset_is_not_false

    def test_vtimezone_with_only_daylight_does_not_raise_assertion(
        self, calendar_with_only_daylight
    ):
        """Test that VTIMEZONE with only DAYLIGHT can be converted to timezone.

        This tests the specific case from issue #321 where a calendar generated
        by khal had a VTIMEZONE with only DAYLIGHT, causing an AssertionError
        when calling to_tz().
        """
        timezone = list(calendar_with_only_daylight.walk())[1]
        assert timezone.name == "VTIMEZONE"
        assert timezone["TZID"] == "Europe/Berlin"

        # This should not raise AssertionError
        # With lookup_tzid=True (default), it finds Europe/Berlin in the system
        tz = timezone.to_tz()
        assert tz is not None

    def test_vtimezone_with_only_daylight_forced_creation(
        self, calendar_with_only_daylight
    ):
        """Test forced timezone creation with lookup_tzid=False.

        This is a more stringent test that forces timezone creation from the
        VTIMEZONE definition rather than looking up an existing timezone.
        This was the specific case that triggered the AssertionError.
        """
        pytest.importorskip("pytz", reason="This test requires pytz")
        from icalendar.timezone.pytz import PYTZ

        pytz_provider = PYTZ()

        timezone = list(calendar_with_only_daylight.walk())[1]

        # Force creation from VTIMEZONE definition
        # This used to fail with AssertionError
        tz = timezone.to_tz(tzp=pytz_provider, lookup_tzid=False)
        assert tz is not None
        assert tz.zone == "Europe/Berlin"



"""Enumerations for different types in the RFCs."""

from enum import Enum as _Enum


class Enum(_Enum):
    """Enum class that can be pickled."""

    def __reduce_ex__(self, _p):
        """For pickling."""
        return self.__class__, (self._name_,)


class StrEnum(str, Enum):
    """Enum for strings."""

    def __str__(self) -> str:
        """Convert to a string.

        This is needed when we set the value directly in components.
        """
        return self.value


class PARTSTAT(StrEnum):
    """Enum for PARTSTAT from :rfc:`5545`.

    Values:
        ``NEEDS_ACTION``,
        ``ACCEPTED``,
        ``DECLINED``,
        ``TENTATIVE``,
        ``DELEGATED``,
        ``COMPLETED``,
        ``IN_PROCESS``

    Purpose:
        To specify the participation status for the calendar user
        specified by the property.

    Description:
        This parameter can be specified on properties with a
        CAL-ADDRESS value type.  The parameter identifies the
        participation status for the calendar user specified by the
        property value.  The parameter values differ depending on whether
        they are associated with a group-scheduled "VEVENT", "VTODO", or
        "VJOURNAL".  The values MUST match one of the values allowed for
        the given calendar component.  If not specified on a property that
        allows this parameter, the default value is NEEDS-ACTION.
        Applications MUST treat x-name and iana-token values they don't
        recognize the same way as they would the NEEDS-ACTION value.
    """

    NEEDS_ACTION = "NEEDS-ACTION"
    ACCEPTED = "ACCEPTED"
    DECLINED = "DECLINED"
    TENTATIVE = "TENTATIVE"
    DELEGATED = "DELEGATED"
    COMPLETED = "COMPLETED"
    IN_PROCESS = "IN-PROCESS"


class STATUS(StrEnum):
    """Enum for STATUS from :rfc:`5545`.

    Values for :class:`icalendar.cal.event.Event`:
        ``CONFIRMED``,
        ``TENTATIVE``,
        ``CANCELLED``

    Values for :class:`icalendar.cal.todo.Todo`:
        ``NEEDS_ACTION``,
        ``COMPLETED``,
        ``IN_PROCESS``,
        ``CANCELLED``

    Values for :class:`icalendar.cal.journal.Journal`:
        ``DRAFT``,
        ``FINAL``,
        ``CANCELLED``

    Description:
        In a group-scheduled calendar component, the property
        is used by the "Organizer" to provide a confirmation of the event
        to the "Attendees".  For example in a "VEVENT" calendar component,
        the "Organizer" can indicate that a meeting is tentative,
        confirmed, or cancelled.  In a "VTODO" calendar component, the
        "Organizer" can indicate that an action item needs action, is
        completed, is in process or being worked on, or has been
        cancelled.  In a "VJOURNAL" calendar component, the "Organizer"
        can indicate that a journal entry is draft, final, or has been
        cancelled or removed.
    """

    # Event
    TENTATIVE = "TENTATIVE"
    CONFIRMED = "CONFIRMED"
    CANCELLED = "CANCELLED"

    # VTodo
    NEEDS_ACTION = "NEEDS-ACTION"
    COMPLETED = "COMPLETED"
    IN_PROCESS = "IN-PROCESS"
    # CANCELLED

    # Journal
    DRAFT = "DRAFT"
    FINAL = "FINAL"
    # CANCELLED


class FBTYPE(StrEnum):
    """Enum for FBTYPE from :rfc:`5545`.

    Values:
        ``FREE``,
        ``BUSY``,
        ``BUSY_UNAVAILABLE``,
        ``BUSY_TENTATIVE``

    See also :class:`BUSYTYPE`.

    Purpose:
        To specify the free or busy time type.

    Description:
        This parameter specifies the free or busy time type.
        The value FREE indicates that the time interval is free for
        scheduling.  The value BUSY indicates that the time interval is
        busy because one or more events have been scheduled for that
        interval.  The value BUSY-UNAVAILABLE indicates that the time
        interval is busy and that the interval can not be scheduled.  The
        value BUSY-TENTATIVE indicates that the time interval is busy
        because one or more events have been tentatively scheduled for
        that interval.  If not specified on a property that allows this
        parameter, the default is BUSY.  Applications MUST treat x-name
        and iana-token values they don't recognize the same way as they
        would the BUSY value.
    """

    FREE = "FREE"
    BUSY = "BUSY"
    BUSY_UNAVAILABLE = "BUSY-UNAVAILABLE"
    BUSY_TENTATIVE = "BUSY-TENTATIVE"


class CUTYPE(StrEnum):
    """Enum for CTYPE from :rfc:`5545`.

    Values:
        ``INDIVIDUAL``,
        ``GROUP``,
        ``RESOURCE``,
        ``ROOM``,
        ``UNKNOWN``

    Purpose:
        To identify the type of calendar user specified by the property.

    Description:
        This parameter can be specified on properties with a
        CAL-ADDRESS value type.  The parameter identifies the type of
        calendar user specified by the property.  If not specified on a
        property that allows this parameter, the default is INDIVIDUAL.
        Applications MUST treat x-name and iana-token values they don't
        recognize the same way as they would the UNKNOWN value.
    """

    INDIVIDUAL = "INDIVIDUAL"
    GROUP = "GROUP"
    RESOURCE = "RESOURCE"
    ROOM = "ROOM"
    UNKNOWN = "UNKNOWN"


class RELTYPE(StrEnum):
    """Enum for RELTYPE from :rfc:`5545`.

    Values:
        ``PARENT``,
        ``CHILD``,
        ``SIBLING``,
        ``FINISHTOSTART``,
        ``FINISHTOFINISH``,
        ``STARTTOFINISH``,
        ``STARTTOSTART``,
        ``FIRST``,
        ``NEXT``,
        ``DEPENDS-ON``,
        ``REFID``,
        ``CONCEPT``

    Purpose:
        To specify the type of hierarchical relationship associated
        with the calendar component specified by the property.

    Description:
        This parameter can be specified on a property that
        references another related calendar.  The parameter specifies the
        hierarchical relationship type of the calendar component
        referenced by the property.  The parameter value can be PARENT, to
        indicate that the referenced calendar component is a superior of
        calendar component; CHILD to indicate that the referenced calendar
        component is a subordinate of the calendar component; or SIBLING
        to indicate that the referenced calendar component is a peer of
        the calendar component.  If this parameter is not specified on an
        allowable property, the default relationship type is PARENT.
        Applications MUST treat x-name and iana-token values they don't
        recognize the same way as they would the PARENT value.

    New Temporal RELTYPE Parameter Values (:rfc:`9253`):
        RELTYPE=FINISHTOSTART
            Task-B cannot start until Task-A finishes.
            For example, when painting is complete, carpet laying can begin.
        RELTYPE=FINISHTOFINISH
            Task-B can only be completed after Task-A is finished.
            The related tasks may run in parallel before completion.
        RELTYPE=STARTTOFINISH
            The start of Task-A (which occurs after Task-B) controls the
            finish of Task-B. For example, ticket sales (Task-B) end
            after the game starts (Task-A).
        RELTYPE=STARTTOSTART
            The start of Task-A triggers the start of Task-B,
            that is, Task-B can start anytime after Task-A starts.

    Additional New RELTYPE Parameter Values (:rfc:`9253`):
        RELTYPE=FIRST
            This indicates that the referenced calendar component is
            the first in a series the referencing calendar component
            is part of.
        RELTYPE=NEXT
            This indicates that the referenced calendar component
            is the next in a series the referencing calendar component
            is part of.
        RELTYPE=DEPENDS-ON
            This indicates that the current calendar component
            depends on the referenced calendar component in some manner.
            For example, a task may be blocked waiting on the other,
            referenced, task.
        RELTYPE=REFID
            This establishes a reference from the current component to
            components with a REFID property that matches the value given
            in the associated RELATED-TO property.
        RELTYPE=CONCEPT
            This establishes a reference from the current component to
            components with a CONCEPT property that matches the value
            given in the associated RELATED-TO property.

        Note that the relationship types of PARENT, CHILD, and SIBLING
        establish a hierarchical relationship. The new types of FIRST
        and NEXT are an ordering relationship.

    Examples:
        The following is an example from :rfc:`5545`:

        .. code-block:: text

            RELATED-TO;RELTYPE=SIBLING:19960401-080045-4000F192713@
             example.com

        Usage inside of RELATED-TO according to :rfc:`9253`:

        .. code-block:: text

            RELATED-TO;VALUE=URI;RELTYPE=STARTTOFINISH:
             https://example.com/caldav/user/jb/cal/
             19960401-080045-4000F192713.ics
    """

    PARENT = "PARENT"
    CHILD = "CHILD"
    SIBLING = "SIBLING"
    FINISHTOSTART = "FINISHTOSTART"
    FINISHTOFINISH = "FINISHTOFINISH"
    STARTTOFINISH = "STARTTOFINISH"
    STARTTOSTART = "STARTTOSTART"
    FIRST = "FIRST"
    NEXT = "NEXT"
    DEPENDS_ON = "DEPENDS-ON"
    REFID = "REFID"
    CONCEPT = "CONCEPT"


class RANGE(StrEnum):
    """Enum for RANGE from :rfc:`5545`.

    Values:
        ``THISANDFUTURE``,
        ``THISANDPRIOR`` (deprecated)

    Purpose:
        To specify the effective range of recurrence instances from
        the instance specified by the recurrence identifier specified by
        the property.

    Description:
        This parameter can be specified on a property that
        specifies a recurrence identifier.  The parameter specifies the
        effective range of recurrence instances that is specified by the
        property.  The effective range is from the recurrence identifier
        specified by the property.  If this parameter is not specified on
        an allowed property, then the default range is the single instance
        specified by the recurrence identifier value of the property.  The
        parameter value can only be "THISANDFUTURE" to indicate a range
        defined by the recurrence identifier and all subsequent instances.
        The value "THISANDPRIOR" is deprecated by this revision of
        iCalendar and MUST NOT be generated by applications.
    """

    THISANDFUTURE = "THISANDFUTURE"
    THISANDPRIOR = "THISANDPRIOR"  # deprecated


class RELATED(StrEnum):
    """Enum for RELATED from :rfc:`5545`.

    Values:
        ``START``,
        ``END``

    Purpose:
        To specify the relationship of the alarm trigger with
        respect to the start or end of the calendar component.

    Description:
        This parameter can be specified on properties that
        specify an alarm trigger with a "DURATION" value type.  The
        parameter specifies whether the alarm will trigger relative to the
        start or end of the calendar component.  The parameter value START
        will set the alarm to trigger off the start of the calendar
        component; the parameter value END will set the alarm to trigger
        off the end of the calendar component.  If the parameter is not
        specified on an allowable property, then the default is START.
    """

    START = "START"
    END = "END"


class ROLE(StrEnum):
    """Enum for ROLE from :rfc:`5545`.

    Values:
        ``CHAIR``,
        ``REQ_PARTICIPANT``,
        ``OPT_PARTICIPANT``,
        ``NON_PARTICIPANT``

    Purpose:
        To specify the participation role for the calendar user
        specified by the property.

    Description:
        This parameter can be specified on properties with a
        CAL-ADDRESS value type.  The parameter specifies the participation
        role for the calendar user specified by the property in the group
        schedule calendar component.  If not specified on a property that
        allows this parameter, the default value is REQ-PARTICIPANT.
        Applications MUST treat x-name and iana-token values they don't
        recognize the same way as they would the REQ-PARTICIPANT value.

    """

    CHAIR = "CHAIR"
    REQ_PARTICIPANT = "REQ-PARTICIPANT"
    OPT_PARTICIPANT = "OPT-PARTICIPANT"
    NON_PARTICIPANT = "NON-PARTICIPANT"


class VALUE(StrEnum):
    """VALUE datatypes as defined in :rfc:`5545`.

    Attributes: ``BOOLEAN``, ``CAL_ADDRESS``, ``DATE``, ``DATE_TIME``, ``DURATION``,
    ``FLOAT``, ``INTEGER``, ``PERIOD``, ``RECUR``, ``TEXT``, ``TIME``, ``URI``,
    ``UTC_OFFSET``

    Description:
        This parameter specifies the value type and format of
        the property value.  The property values MUST be of a single value
        type.  For example, a "RDATE" property cannot have a combination
        of DATE-TIME and TIME value types.

        If the property's value is the default value type, then this
        parameter need not be specified.  However, if the property's
        default value type is overridden by some other allowable value
        type, then this parameter MUST be specified.

        Applications MUST preserve the value data for x-name and iana-
        token values that they don't recognize without attempting to
        interpret or parse the value data.

    """

    BOOLEAN = "BOOLEAN"
    CAL_ADDRESS = "CAL-ADDRESS"
    DATE = "DATE"
    DATE_TIME = "DATE-TIME"
    DURATION = "DURATION"
    FLOAT = "FLOAT"
    INTEGER = "INTEGER"
    PERIOD = "PERIOD"
    RECUR = "RECUR"
    TEXT = "TEXT"
    TIME = "TIME"
    URI = "URI"
    UTC_OFFSET = "UTC-OFFSET"


class BUSYTYPE(StrEnum):
    """Enum for BUSYTYPE from :rfc:`7953`.

    Values:
        ``BUSY``,
        ``BUSY_UNAVAILABLE``,
        ``BUSY_TENTATIVE``

    Description:
        This property is used to specify the default busy time
        type.  The values correspond to those used by the :class:`FBTYPE`
        parameter used on a "FREEBUSY" property, with the exception that
        the "FREE" value is not used in this property.  If not specified
        on a component that allows this property, the default is "BUSY-
        UNAVAILABLE".

    Example:
        The following is an example of this property:

        .. code-block:: text

            BUSYTYPE:BUSY
    """

    BUSY = "BUSY"
    BUSY_UNAVAILABLE = "BUSY-UNAVAILABLE"
    BUSY_TENTATIVE = "BUSY-TENTATIVE"


class CLASS(StrEnum):
    """Enum for CLASS from :rfc:`5545`.

    Values:
        ``PUBLIC``,
        ``PRIVATE``,
        ``CONFIDENTIAL``

    Description:
        An access classification is only one component of the
        general security system within a calendar application.  It
        provides a method of capturing the scope of the access the
        calendar owner intends for information within an individual
        calendar entry.  The access classification of an individual
        iCalendar component is useful when measured along with the other
        security components of a calendar system (e.g., calendar user
        authentication, authorization, access rights, access role, etc.).
        Hence, the semantics of the individual access classifications
        cannot be completely defined by this memo alone.  Additionally,
        due to the "blind" nature of most exchange processes using this
        memo, these access classifications cannot serve as an enforcement
        statement for a system receiving an iCalendar object.  Rather,
        they provide a method for capturing the intention of the calendar
        owner for the access to the calendar component.  If not specified
        in a component that allows this property, the default value is
        PUBLIC.  Applications MUST treat x-name and iana-token values they
        don't recognize the same way as they would the PRIVATE value.
    """

    PUBLIC = "PUBLIC"
    PRIVATE = "PRIVATE"
    CONFIDENTIAL = "CONFIDENTIAL"


class TRANSP(StrEnum):
    """Enum for TRANSP from :rfc:`5545`.

    Values:
        ``OPAQUE``,
        ``TRANSPARENT``

    Description:
        Time Transparency is the characteristic of an event
        that determines whether it appears to consume time on a calendar.
        Events that consume actual time for the individual or resource
        associated with the calendar SHOULD be recorded as OPAQUE,
        allowing them to be detected by free/busy time searches.  Other
        events, which do not take up the individual's (or resource's) time
        SHOULD be recorded as TRANSPARENT, making them invisible to free/
        busy time searches.
    """

    OPAQUE = "OPAQUE"
    TRANSPARENT = "TRANSPARENT"


__all__ = [
    "BUSYTYPE",
    "CLASS",
    "CUTYPE",
    "FBTYPE",
    "PARTSTAT",
    "RANGE",
    "RELATED",
    "RELTYPE",
    "ROLE",
    "STATUS",
    "TRANSP",
    "VALUE",
]



"""Tests for issue #867: Todo.duration should work without DTSTART."""

from datetime import datetime, timedelta

import pytest

from icalendar import Event, Todo
from icalendar.error import IncompleteComponent


def test_todo_duration_without_dtstart():
    """Test that Todo.duration works when `DURATION` property is set but `DTSTART` is missing."""
    # This is the exact test case from issue #867
    my_task = Todo.from_ical("""BEGIN:VTODO
UID:taskwithoutdtstart
DTSTAMP:20070313T123432Z
DURATION:P5D
SUMMARY:This is a task that is expected to take five days to complete
STATUS:NEEDS-ACTION
END:VTODO""")

    # Should return the DURATION property directly
    assert my_task.duration == timedelta(days=5)


def test_todo_duration_calculated_from_start_and_due():
    """Test that Todo.duration still works for calculated duration from `DTSTART` and `DUE`."""
    todo = Todo()
    todo.add("UID", "test-calculated")
    todo.start = datetime(2026, 3, 19, 12, 0)
    todo.end = datetime(2026, 3, 19, 15, 30)

    # Should calculate duration from start and end
    assert todo.duration == timedelta(hours=3, minutes=30)


def test_todo_duration_prefers_duration_property():
    """Test that explicit `DURATION` property takes precedence over calculated duration."""
    todo = Todo()
    todo.add("UID", "test-precedence")
    todo.start = datetime(2026, 3, 19, 12, 0)
    todo.end = datetime(2026, 3, 19, 15, 0)  # This would be 3 hours
    todo.add("DURATION", timedelta(days=2))  # But DURATION says 2 days

    # Should return DURATION property, not calculated value
    assert todo.duration == timedelta(days=2)


def test_todo_duration_with_dtstart_and_duration():
    """Test Todo with `DTSTART` and `DURATION` (valid per RFC 5545)."""
    todo = Todo()
    todo.add("UID", "test-start-duration")
    todo.start = datetime(2026, 3, 19, 12, 0)
    todo.add("DURATION", timedelta(hours=4))

    # Should return DURATION property
    assert todo.duration == timedelta(hours=4)


def test_todo_duration_without_any_time_info_raises_error():
    """Test that Todo.duration raises error when no time information is available."""
    todo = Todo()
    todo.add("UID", "test-no-time")
    todo.add("SUMMARY", "Task without any time info")

    # Should raise error since no DURATION, DTSTART, or DUE is set
    with pytest.raises(IncompleteComponent):
        _ = todo.duration


def test_todo_duration_complex_duration_values():
    """Test Todo.duration with various complex duration formats."""
    # Test ISO 8601 duration formats
    test_cases = [
        ("PT30M", timedelta(minutes=30)),
        ("PT1H30M", timedelta(hours=1, minutes=30)),
        ("P1DT2H", timedelta(days=1, hours=2)),
        ("P1W", timedelta(weeks=1)),
        (
            "P1Y2M3DT4H5M6S",
            timedelta(days=428, hours=4, minutes=5, seconds=6),
        ),  # Approx 1 year 2 months
    ]

    for duration_str, expected_delta in test_cases:
        todo = Todo()
        todo.add("UID", f"test-{duration_str}")
        todo.add("DURATION", expected_delta)

        assert todo.duration == expected_delta


def test_todo_duration_maintains_backward_compatibility():
    """Test that the fix doesn't break existing functionality."""
    # Create todo the old way (DTSTART + DUE)
    todo = Todo()
    todo.add("UID", "test-backward-compat")
    todo.add("SUMMARY", "Backward compatibility test")
    todo.start = datetime(2026, 3, 19, 9, 0)
    todo.end = datetime(2026, 3, 19, 17, 0)

    # Should still work as before
    assert todo.duration == timedelta(hours=8)

    # Properties should still work
    assert todo.start == datetime(2026, 3, 19, 9, 0)
    assert todo.end == datetime(2026, 3, 19, 17, 0)


def test_todo_duration_edge_case_only_dtstart():
    """Test Todo with only `DTSTART` (no `DUE` or `DURATION`)."""
    todo = Todo()
    todo.add("UID", "test-only-start")
    todo.start = datetime(2026, 3, 19, 12, 0)

    # This should use the fallback logic from the end property
    # which returns start + 1 day for date, or just start for datetime
    assert todo.duration == timedelta(0)  # end defaults to start for datetime


def test_issue_867_exact_reproduction():
    """Test the exact scenario reported in issue #867."""
    # This reproduces the caldav project test case that was failing
    my_task = Todo.from_ical("""BEGIN:VTODO
UID:taskwithoutdtstart
DTSTAMP:20070313T123432Z
DURATION:P5D
SUMMARY:This is a task that is expected to take five days to complete
STATUS:NEEDS-ACTION
END:VTODO""")

    # This assertion was failing before the fix
    assert my_task.duration == timedelta(days=5)

    # Verify the todo has the expected properties
    assert my_task.get("UID") == "taskwithoutdtstart"
    assert (
        my_task.get("SUMMARY")
        == "This is a task that is expected to take five days to complete"
    )
    assert "DURATION" in my_task
    assert "DTSTART" not in my_task  # Confirming no DTSTART


def test_todo_duration_preserves_property_access():
    """Test that direct property access still works as expected."""
    todo = Todo()
    todo.add("UID", "test-property-access")
    todo.add("DURATION", timedelta(hours=2))

    # Direct property access should still work
    assert todo["DURATION"].dt == timedelta(hours=2)

    # And our computed property should match
    assert todo.duration == timedelta(hours=2)


# Parametrized tests for both Event and Todo components
@pytest.mark.parametrize("component_class", [Event, Todo])
def test_component_duration_prefers_duration_property(component_class):
    """Test that both Event and Todo prefer `DURATION` property over calculated duration."""
    component = component_class()
    component.add("UID", f"test-{component_class.__name__.lower()}-duration")
    component.start = datetime(2026, 3, 19, 12, 0)
    component.end = datetime(2026, 3, 19, 15, 0)  # This would be 3 hours
    component.add("DURATION", timedelta(hours=2))  # But DURATION says 2 hours

    # Should return DURATION property, not calculated value
    assert component.duration == timedelta(hours=2)


@pytest.mark.parametrize("component_class", [Event, Todo])
def test_component_duration_calculated_fallback(component_class):
    """Test that both Event and Todo fall back to calculated duration when no `DURATION` property."""
    component = component_class()
    component.add("UID", f"test-{component_class.__name__.lower()}-calculated")
    component.start = datetime(2026, 3, 19, 12, 0)
    component.end = datetime(2026, 3, 19, 14, 30)

    # Should calculate duration from start and end (no DURATION property)
    assert component.duration == timedelta(hours=2, minutes=30)



from icalendar.parser import Parameters
from icalendar.prop import vCalAddress

txt = b"MAILTO:maxm@mxm.dk"
a = vCalAddress(txt)
a.params["cn"] = "Max M"


def test_to_ical():
    assert a.to_ical() == txt


def test_params():
    assert isinstance(a.params, Parameters)
    assert a.params == {"CN": "Max M"}


def test_from_ical():
    assert vCalAddress.from_ical(txt) == "MAILTO:maxm@mxm.dk"


def test_repr():
    instance = vCalAddress("value")
    assert repr(instance) == "vCalAddress('value')"


def test_email_malformed():
    """Sometimes, people forget to add mailto that."""
    address = vCalAddress("me@you.we")
    assert address.email == "me@you.we"


def test_email_mailto():
    """Email with a normal mailto link."""
    address = vCalAddress("mailto:icalendar@email.list")
    assert address.email == "icalendar@email.list"


def test_capital_email():
    """mailto can be capital letters."""
    address = vCalAddress("MAILTO:yemaya@posteo.net")
    assert address.email == "yemaya@posteo.net"


def test_name():
    """We want the name, too!"""
    address = vCalAddress("MAILTO:yemaya@posteo.net")
    assert address.name == ""
    address.params["CN"] = "name!"
    assert address.name == "name!"


def test_set_the_name():
    address = vCalAddress("MAILTO:yemaya@posteo.net")
    address.name = "Yemaya :)"
    assert address.name == "Yemaya :)"
    assert address.params["CN"] == "Yemaya :)"



"""Test the enums for values."""

import pytest

import icalendar
from icalendar import enums

ENUMS = [name for name in dir(enums) if name.isupper()]


@pytest.fixture(params=ENUMS)
def enum_name(request):
    """The name of an enum"""
    return request.param


@pytest.fixture
def enum(enum_name):
    """An enum."""
    return getattr(enums, enum_name)


def test_all_enums_are_exported(enum_name):
    """All enums should be exported."""
    assert enum_name in enums.__all__


def test_all_enums_are_public(enum_name):
    """All enums should be exported."""
    assert enum_name in icalendar.__all__, f"icalendar.__all__ is missing {enum_name}"


def test_enum_has_description(enum):
    """We should have a docstring."""
    assert "Description:" in enum.__doc__



from typing import List, Union

SEQUENCE_TYPES = (list, tuple)
DEFAULT_ENCODING = "utf-8"
ICAL_TYPE = Union[str, bytes]


def from_unicode(value: ICAL_TYPE, encoding="utf-8") -> bytes:
    """
    Converts a value to bytes, even if it already is bytes
    :param value: The value to convert
    :param encoding: The encoding to use in the conversion
    :return: The bytes representation of the value
    """
    if isinstance(value, bytes):
        return value
    elif isinstance(value, str):
        try:
            return value.encode(encoding)
        except UnicodeEncodeError:
            return value.encode("utf-8", "replace")
    else:
        return value


def to_unicode(value: ICAL_TYPE, encoding="utf-8-sig") -> str:
    """Converts a value to unicode, even if it is already a unicode string."""
    if isinstance(value, str):
        return value
    elif isinstance(value, bytes):
        try:
            return value.decode(encoding)
        except UnicodeDecodeError:
            return value.decode("utf-8-sig", "replace")
    else:
        return value


def data_encode(
    data: Union[ICAL_TYPE, dict, list], encoding=DEFAULT_ENCODING
) -> Union[bytes, List[bytes], dict]:
    """Encode all datastructures to the given encoding.
    Currently unicode strings, dicts and lists are supported.
    """
    # https://stackoverflow.com/questions/1254454/fastest-way-to-convert-a-dicts-keys-values-from-unicode-to-str
    if isinstance(data, str):
        return data.encode(encoding)
    elif isinstance(data, dict):
        return dict(map(data_encode, iter(data.items())))
    elif isinstance(data, list) or isinstance(data, tuple):
        return list(map(data_encode, data))
    else:
        return data


__all__ = [
    "DEFAULT_ENCODING",
    "SEQUENCE_TYPES",
    "ICAL_TYPE",
    "data_encode",
    "from_unicode",
    "to_unicode",
]



"""Tests checking that parsing works"""

import base64
from datetime import datetime

import pytest

from icalendar import vBinary, vRecur
from icalendar.cal.calendar import Calendar
from icalendar.cal.component_factory import ComponentFactory
from icalendar.cal.event import Event
from icalendar.parser import Contentline, Parameters, unescape_char


@pytest.mark.parametrize(
    "calendar_name",
    [
        # Issue #178 - A component with an unknown/invalid name is represented
        # as one of the known components, the information about the original
        # component name is lost.
        # https://github.com/collective/icalendar/issues/178 https://github.com/collective/icalendar/pull/180
        # Parsing of a nonstandard component
        "issue_178_component_with_invalid_name_represented",
        # Nonstandard component inside other components, also has properties
        "issue_178_custom_component_inside_other",
        # Nonstandard component is able to contain other components
        "issue_178_custom_component_contains_other",
    ],
)
def test_calendar_to_ical_is_inverse_of_from_ical(calendars, calendar_name):
    calendar = getattr(calendars, calendar_name)
    assert calendar.to_ical().splitlines() == calendar.raw_ics.splitlines()
    assert calendar.to_ical() == calendar.raw_ics


@pytest.mark.parametrize(
    ("raw_content_line", "expected_output"),
    [
        # Issue #142 - Multivalued parameters. This is needed for VCard 3.0.
        # see https://github.com/collective/icalendar/pull/142
        (
            "TEL;TYPE=HOME,VOICE:000000000",
            ("TEL", Parameters({"TYPE": ["HOME", "VOICE"]}), "000000000"),
        ),
        # Issue #143 - Allow dots in property names. Another vCard related issue.
        # see https://github.com/collective/icalendar/pull/143
        (
            "ITEMADRNULLTHISISTHEADRESS08158SOMECITY12345.ADR:;;This is the Adress 08; Some City;;12345;Germany",
            (
                "ITEMADRNULLTHISISTHEADRESS08158SOMECITY12345.ADR",
                Parameters(),
                ";;This is the Adress 08; Some City;;12345;Germany",
            ),
        ),
        (
            "ITEMADRNULLTHISISTHEADRESS08158SOMECITY12345.X-ABLABEL:",
            (
                "ITEMADRNULLTHISISTHEADRESS08158SOMECITY12345.X-ABLABEL",
                Parameters(),
                "",
            ),
        ),
    ],
)
def test_content_lines_parsed_properly(raw_content_line, expected_output):
    assert Contentline.from_ical(raw_content_line).parts() == expected_output


@pytest.mark.parametrize(
    "timezone_info",
    [
        # General timezone aware dates in ical string
        (b"DTSTART;TZID=America/New_York:20130907T120000"),
        (b"DTEND;TZID=America/New_York:20130907T170000"),
        # Specific timezone aware exdates in ical string
        (b"EXDATE;TZID=America/New_York:20131012T120000"),
        (b"EXDATE;TZID=America/New_York:20131011T120000"),
    ],
)
def test_timezone_info_present_in_ical_issue_112(events, timezone_info):
    """Issue #112 - No timezone info on EXDATE

    https://github.com/collective/icalendar/issues/112
    """
    assert timezone_info in events.issue_112_missing_tzinfo_on_exdate.to_ical()


def test_timezone_name_parsed_issue_112(events):
    """Issue #112 - No timezone info on EXDATE

    https://github.com/collective/icalendar/issues/112
    """
    assert (
        events.issue_112_missing_tzinfo_on_exdate["exdate"][0].dts[0].dt.tzname()
        == "EDT"
    )


def test_issue_157_removes_trailing_semicolon(events):
    """Issue #157 - Recurring rules and trailing semicolons

    https://github.com/collective/icalendar/pull/157
    """
    recur = events.issue_157_removes_trailing_semicolon.decoded("RRULE")
    assert isinstance(recur, vRecur)
    assert recur.to_ical() == b"FREQ=YEARLY;BYDAY=1SU;BYMONTH=11"


@pytest.mark.parametrize(
    "event_name",
    [
        # https://github.com/collective/icalendar/pull/100
        ("issue_100_transformed_doctests_into_unittests"),
        ("issue_184_broken_representation_of_period"),
        # PERIOD should be put back into shape
        "issue_156_RDATE_with_PERIOD",
        "issue_156_RDATE_with_PERIOD_list",
        "event_with_unicode_organizer",
    ],
)
def test_event_to_ical_is_inverse_of_from_ical(events, event_name):
    """Make sure that an event's ICS is equal to the ICS it was made from."""
    event = events[event_name]
    assert event.to_ical().splitlines() == event.raw_ics.splitlines()
    assert event.to_ical() == event.raw_ics


def test_decode_rrule_attribute_error_issue_70(events):
    # Issue #70 - e.decode("RRULE") causes Attribute Error
    # see https://github.com/collective/icalendar/issues/70
    recur = events.issue_70_rrule_causes_attribute_error.decoded("RRULE")
    assert isinstance(recur, vRecur)
    assert recur.to_ical() == b"FREQ=WEEKLY;UNTIL=20070619T225959;INTERVAL=1"


def test_description_parsed_properly_issue_53(events):
    """Issue #53 - Parsing failure on some descriptions?

    https://github.com/collective/icalendar/issues/53
    """
    assert (
        b"July 12 at 6:30 PM"
        in events.issue_53_description_parsed_properly["DESCRIPTION"].to_ical()
    )


def test_raises_value_error_for_properties_without_parent_pull_179():
    """Found an issue where from_ical() would raise IndexError for
    properties without parent components.

    https://github.com/collective/icalendar/pull/179
    """
    with pytest.raises(ValueError):
        Calendar.from_ical("VERSION:2.0")


def test_tzid_parsed_properly_issue_53(timezones):
    """Issue #53 - Parsing failure on some descriptions?

    https://github.com/collective/icalendar/issues/53
    """
    assert (
        timezones.issue_53_tzid_parsed_properly["tzid"].to_ical() == b"America/New_York"
    )


def test_timezones_to_ical_is_inverse_of_from_ical(timezones):
    """Issue #55 - Parse error on utc-offset with seconds value
    see https://github.com/collective/icalendar/issues/55"""
    timezone = timezones["issue_55_parse_error_on_utc_offset_with_seconds"]
    assert timezone.to_ical() == timezone.raw_ics


@pytest.mark.parametrize(
    ("date", "expected_output"),
    [
        (datetime(2012, 7, 16, 0, 0, 0), b"DTSTART:20120716T000000Z"),
        (datetime(2021, 11, 17, 15, 9, 15), b"DTSTART:20211117T150915Z"),
    ],
)
def test_no_tzid_when_utc(utc, date, expected_output):
    """Issue #58  - TZID on UTC DATE-TIMEs
       Issue #335 - UTC timezone identification is broken

    https://github.com/collective/icalendar/issues/58
    https://github.com/collective/icalendar/issues/335
    """
    # According to RFC 5545: "The TZID property parameter MUST NOT be
    # applied to DATE-TIME or TIME properties whose time values are
    # specified in UTC.
    date = date.replace(tzinfo=utc)
    event = Event()
    event.add("dtstart", date)
    assert expected_output in event.to_ical()


def test_vBinary_base64_encoded_issue_82():
    """Issue #82 - vBinary __repr__ called rather than to_ical from
                   container types
    https://github.com/collective/icalendar/issues/82
    """
    b = vBinary("text")
    b.params["FMTTYPE"] = "text/plain"
    assert b.to_ical() == base64.b64encode(b"text")


def test_creates_event_with_base64_encoded_attachment_issue_82(events):
    """Issue #82 - vBinary __repr__ called rather than to_ical from
                   container types
    https://github.com/collective/icalendar/issues/82
    """
    b = vBinary("text")
    b.params["FMTTYPE"] = "text/plain"
    event = Event()
    event.add("ATTACH", b)
    assert event.to_ical() == events.issue_82_expected_output.raw_ics


@pytest.mark.parametrize(
    "calendar_name",
    [
        # Issue #466 - [BUG] TZID timezone is ignored when forward-slash is used
        # https://github.com/collective/icalendar/issues/466
        "issue_466_respect_unique_timezone",
        "issue_466_convert_tzid_with_slash",
    ],
)
def test_handles_unique_tzid(calendars, in_timezone, calendar_name):
    calendar = calendars[calendar_name]
    event = calendar.walk("VEVENT")[0]
    print(vars(event))
    start_dt = event["dtstart"].dt
    end_dt = event["dtend"].dt
    assert start_dt == in_timezone(datetime(2022, 10, 21, 20, 0, 0), "Europe/Stockholm")
    assert end_dt == in_timezone(datetime(2022, 10, 21, 21, 0, 0), "Europe/Stockholm")


@pytest.mark.parametrize(
    ("event_name", "expected_cn", "expected_ics"),
    [
        (
            "event_with_escaped_characters",
            r"that, that; %th%%at%\ that:",
            "это, то; that\\ %th%%at%:",
        ),
        ("event_with_escaped_character1", r"Society, 2014", "that"),
        ("event_with_escaped_character2", r"Society\ 2014", "that"),
        ("event_with_escaped_character3", r"Society; 2014", "that"),
        ("event_with_escaped_character4", r"Society: 2014", "that"),
    ],
)
def test_escaped_characters_read(event_name, expected_cn, expected_ics, events):
    event = events[event_name]
    assert event["ORGANIZER"].params["CN"] == expected_cn
    assert event["ORGANIZER"].to_ical() == expected_ics.encode("utf-8")


def test_unescape_char():
    assert unescape_char(b"123") == b"123"
    assert unescape_char(b"\\n") == b"\n"


def test_create_a_component():
    """Create a component with the factory."""
    factory = ComponentFactory()
    my_component_class = factory.get_component_class("My-Component")
    assert my_component_class.name == "MY-COMPONENT"
    assert my_component_class.__name__ == "MyComponent"



"""Check serialization and deserialization of values.

See https://datatracker.ietf.org/doc/html/rfc7265#section-3.6
"""

from datetime import date, datetime, time, timedelta

import pytest

from icalendar import (
    Calendar,
    vBinary,
    vBoolean,
    vCalAddress,
    vDate,
    vDatetime,
    vDDDLists,
    vDDDTypes,
    vDuration,
    vFloat,
    vInt,
    vPeriod,
    vRecur,
    vText,
    vTime,
    vUnknown,
    vUri,
    vUTCOffset,
)
from icalendar.compatibility import ZoneInfo

JCAL_PAIRS = [
    (["attach", {}, "binary", "SGVsbG8gV29ybGQh"], vBinary("SGVsbG8gV29ybGQh")),
    (["x-non-smoking", {}, "boolean", True], vBoolean(True)),  # noqa: FBT003
    (["x-non-smoking", {}, "boolean", False], vBoolean(False)),  # noqa: FBT003
    (
        ["attendee", {}, "cal-address", "mailto:kewisch@example.com"],
        vCalAddress("mailto:kewisch@example.com"),
    ),
    (["dtstart", {}, "date", "2011-05-17"], vDate(date(2011, 5, 17))),
    (
        ["dtstart", {}, "date-time", "2012-10-17T12:00:00"],
        vDatetime(datetime(2012, 10, 17, 12, 0, 0)),
    ),
    (
        ["dtstamp", {}, "date-time", "2012-11-17T12:00:00Z"],
        vDatetime(datetime(2012, 11, 17, 12, 0, 0, tzinfo=ZoneInfo("UTC"))),
    ),
    (
        ["dtend", {"tzid": "Europe/Berlin"}, "date-time", "2011-12-17T13:00:00"],
        vDatetime(datetime(2011, 12, 17, 13, 0, 0, tzinfo=ZoneInfo("Europe/Berlin"))),
    ),
    (["duration", {}, "duration", "P1D"], vDuration(timedelta(days=1))),
    (["x-grade", {}, "float", 1.3], vFloat(1.3)),
    (["percent-complete", {}, "integer", 42], vInt(42)),
    (["x-foo", {}, "integer", -42], vInt(-42)),
    (
        ["freebusy", {"fbtype": "FREE"}, "period", ["1997-03-08T16:00:00Z", "P1D"]],
        vPeriod(
            (datetime(1997, 3, 8, 16, 0, 0, tzinfo=ZoneInfo("UTC")), timedelta(days=1)),
            params={"fbtype": "FREE"},
        ),
    ),
    (
        [
            "freebusy",
            {"fbtype": "FREE", "tzid": "Europe/Moscow"},
            "period",
            ["1997-03-08T16:00:00", "P1D"],
        ],
        vPeriod(
            (
                datetime(1997, 3, 8, 16, 0, 0, tzinfo=ZoneInfo("Europe/Moscow")),
                timedelta(days=1),
            ),
            params={"fbtype": "FREE"},
        ),
    ),
    (
        [
            "freebusy",
            {"fbtype": "FREE"},
            "period",
            ["1997-03-08T16:00:00", "1997-03-08T16:00:01"],
        ],
        vPeriod(
            (datetime(1997, 3, 8, 16, 0, 0), datetime(1997, 3, 8, 16, 0, 1)),
            params={"fbtype": "FREE"},
        ),
    ),
    (
        [
            "rrule",
            {},
            "recur",
            {"freq": "YEARLY", "count": 5, "byday": ["-1SU", "2MO"], "bymonth": [10]},
        ],
        vRecur({"freq": "YEARLY", "count": 5, "bymonth": 10, "byday": ["-1SU", "2MO"]}),
    ),
    (
        [
            "rrule",
            {},
            "recur",
            {
                "freq": "MONTHLY",
                "interval": 2,
                "bymonthday": [1, 15, -1],
                "until": "2013-10-01",
            },
        ],
        vRecur(
            {
                "freq": "MONTHLY",
                "interval": 2,
                "bymonthday": [1, 15, -1],
                "until": [date(2013, 10, 1)],
            }
        ),
    ),
    (["comment", {}, "text", "hello, world"], vText("hello, world")),
    (["x-time-local", {}, "time", "12:30:00"], vTime(time(hour=12, minute=30))),
    (
        ["x-time-utc", {}, "time", "12:31:00Z"],
        vTime(time(hour=12, minute=31, tzinfo=ZoneInfo("UTC"))),
    ),
    (
        ["x-time-offset", {"tzid": "Europe/Berlin"}, "time", "12:30:00"],
        vTime(time(hour=12, minute=30, tzinfo=ZoneInfo("Europe/Berlin"))),
    ),
    (
        ["tzurl", {}, "uri", "http://example.org/tz/Europe-Berlin.ics"],
        vUri("http://example.org/tz/Europe-Berlin.ics"),
    ),
    (["tzoffsetfrom", {}, "utc-offset", "-05:00"], vUTCOffset(timedelta(hours=-5))),
    (
        ["tzoffsetto", {}, "utc-offset", "+12:45"],
        vUTCOffset(timedelta(hours=12, minutes=45)),
    ),
    (["x-foo", {}, "unknown", "bar"], vUnknown("bar")),
    (
        ["trigger", {}, "date-time", "1976-04-01T00:55:45Z"],
        vDDDTypes(datetime(1976, 4, 1, 0, 55, 45, tzinfo=ZoneInfo("UTC"))),
    ),
]

mark_pairs = pytest.mark.parametrize(("jcal_value", "prop"), JCAL_PAIRS)


@mark_pairs
def test_convert_property_to_jcal(jcal_value, prop):
    """Check converting an icalendar property type into a jCal value"""
    assert prop.to_jcal(jcal_value[0]) == jcal_value


@mark_pairs
def test_convert_jcal_to_property(jcal_value, prop):
    """Check converting an icalendar property type into a jCal value"""
    assert prop.from_jcal(jcal_value) == prop


vDDD_pairs = [  # noqa: N816
    (jcal, prop)
    for jcal, prop in JCAL_PAIRS
    if isinstance(prop, (vDatetime, vDate, vTime, vDuration))
]


@pytest.mark.parametrize(("jcal_value", "prop"), vDDD_pairs)
@pytest.mark.parametrize("ddd_type", [vDDDTypes, vDDDLists])
def test_parse_from_ddd(ddd_type, jcal_value, prop):
    """The DDD types should be able to parse from a jCal value that is of a supported type."""
    parsed = ddd_type.from_jcal(jcal_value)
    print(parsed)
    print(prop, "expected")
    assert parsed == prop


def test_adding_unknown_value_parameter():
    """When converting to iCalendar, [...] the VALUE parameter
    MUST be omitted for properties that have the jCal type identifier
    "unknown"."""
    jcal = ["x-foo", {}, "unknown", "bar"]
    prop = vUnknown.from_jcal(jcal)
    calendar = Calendar()
    calendar["X-FOO"] = prop
    ical = calendar.to_ical().decode("utf-8")
    assert "UNKNOWN" not in ical
    assert "VALUE" not in ical



"""Test the access to parameters via getter/setter properties."""

import pytest

from icalendar import Parameters
from icalendar.enums import VALUE


@pytest.fixture
def p():
    """Empty test property."""
    return Parameters()


def test_value_parameter_default(p):
    """Test default value."""
    assert p.value is None


def test_set_value(p):
    """Set the value, convert"""
    p.value = "test"
    assert p.value == "TEST"
    assert p["VALUE"] == "TEST"


def test_conversion(p):
    """Converted to enum."""
    p.value = "DATE-TIME"
    assert p.value == VALUE.DATE_TIME


def test_delete_value(p):
    """Delete the value."""
    p.value = "test"
    del p.value
    assert p.value is None
    assert "VALUE" not in p
    del p.value
    assert p.value is None
    assert "VALUE" not in p


def test_delete_value_None(p):
    """Delete the value."""
    p.value = "test"
    p.value = None
    assert p.value is None
    assert "VALUE" not in p
    p.value = None
    assert p.value is None
    assert "VALUE" not in p



"""This module helps identifying the timezone ids and where they differ.

The algorithm: We use the tzname and the utcoffset for each hour from
1970 - 2030.
We make a big map.
If they are equivalent, they are equivalent within the time that is mostly used.

You can regenerate the information from this module.

See also:
- https://stackoverflow.com/questions/79171631/how-do-i-determine-whether-a-zoneinfo-is-an-alias/79171734#79171734

Run this module:

    python -m icalendar.timezone.equivalent_timezone_ids

"""

from __future__ import annotations

from collections import defaultdict
from datetime import datetime, timedelta, tzinfo
from pathlib import Path
from pprint import pprint
from typing import Callable, NamedTuple, Optional

from pytz import AmbiguousTimeError, NonExistentTimeError
from zoneinfo import ZoneInfo, available_timezones

START = datetime(1970, 1, 1)  # noqa: DTZ001
END = datetime(2020, 1, 1)  # noqa: DTZ001
DISTANCE_FROM_TIMEZONE_CHANGE = timedelta(hours=12)

DTS = []
dt = START
while dt <= END:
    DTS.append(dt)
    dt += timedelta(
        hours=25
    )  # This must be big enough to be fast and small enough to identify the timeszones before it is the present year
del dt


def main(
    create_timezones: list[Callable[[str], tzinfo]],
    name: str,
):
    """Generate a lookup table for timezone information if unknown timezones.

    We cannot create one lookup for all because they seem to be all equivalent
    if we mix timezone implementations.
    """
    print(create_timezones, name)
    unsorted_tzids = available_timezones()
    unsorted_tzids.remove("localtime")
    unsorted_tzids.remove("Factory")

    class TZ(NamedTuple):
        tz: tzinfo
        id: str

    tzs = [
        TZ(create_timezone(tzid), tzid)
        for create_timezone in create_timezones
        for tzid in unsorted_tzids
    ]

    def generate_tree(
        tzs: list[TZ],
        step: timedelta = timedelta(hours=1),
        start: datetime = START,
        end: datetime = END,
        todo: Optional[set[str]] = None,
    ) -> tuple[datetime, dict[timedelta, set[str]]] | set[str]:  # should be recursive
        """Generate a lookup tree."""
        if todo is None:
            todo = [tz.id for tz in tzs]
        print(f"{len(todo)} left to compute")
        print(len(tzs))
        if len(tzs) == 0:
            raise ValueError("tzs cannot be empty")
        if len(tzs) == 1:
            todo.remove(tzs[0].id)
            return {tzs[0].id}
        while start < end:
            offsets: dict[timedelta, list[TZ]] = defaultdict(list)
            try:
                # if we are around a timezone change, we must move on
                # see https://github.com/collective/icalendar/issues/776
                around_tz_change = not all(
                    tz.tz.utcoffset(start)
                    == tz.tz.utcoffset(start - DISTANCE_FROM_TIMEZONE_CHANGE)
                    == tz.tz.utcoffset(start + DISTANCE_FROM_TIMEZONE_CHANGE)
                    for tz in tzs
                )
            except (NonExistentTimeError, AmbiguousTimeError):
                around_tz_change = True
            if around_tz_change:
                start += DISTANCE_FROM_TIMEZONE_CHANGE
                continue
            for tz in tzs:
                offsets[tz.tz.utcoffset(start)].append(tz)
            if len(offsets) == 1:
                start += step
                continue
            lookup = {}
            for offset, tzs in offsets.items():
                lookup[offset] = generate_tree(
                    tzs=tzs, step=step, start=start + step, end=end, todo=todo
                )
            return start, lookup
        print(f"reached end with {len(tzs)} timezones - assuming they are equivalent.")
        result = set()
        for tz in tzs:
            result.add(tz.id)
            todo.remove(tz.id)
        return result

    lookup = generate_tree(tzs, step=timedelta(hours=33))

    file = Path(__file__).parent / f"equivalent_timezone_ids_{name}.py"
    print(f"The result is written to {file}.")
    print("lookup = ", end="")
    pprint(lookup)
    with file.open("w") as f:
        f.write(
            f"'''This file is automatically generated by {Path(__file__).name}'''\n"
        )
        f.write("import datetime\n\n")
        f.write("\nlookup = ")
        pprint(lookup, stream=f)
        f.write("\n\n__all__ = ['lookup']\n")

    return lookup


__all__ = ["main"]

if __name__ == "__main__":
    from dateutil.tz import gettz
    from pytz import timezone
    from zoneinfo import ZoneInfo

    # add more timezone implementations if you like
    main(
        [ZoneInfo, timezone, gettz],
        "result",
    )



"""We would like to have certain errors for invalid JCal."""

from json import JSONDecodeError

import pytest

from icalendar import (
    Calendar,
    Component,
    JCalParsingError,
    Parameters,
    vDate,
    vDatetime,
    vDDDTypes,
    vDuration,
    vPeriod,
    vTime,
)
from icalendar.prop import vCategory, vMonth, vRecur, vSkip


def test_invalid_json():
    """JSON can be invalid"""
    with pytest.raises(JSONDecodeError):
        Component.from_jcal("")


@pytest.fixture(params=[[], {}, None, 1, 1.3])
def str_expected(request):
    """Return everything else"""
    return request.param


@pytest.fixture(params=["a", {}, None, 1, 1.3])
def list_expected(request):
    """Return everything else"""
    return request.param


@pytest.fixture(params=["a", [], None, 1, 1.3])
def object_expected(request):
    """Return everything else"""
    return request.param


@pytest.fixture(params=["a", [], {}, None, 1.3])
def int_expected(request):
    """Return everything else"""
    return request.param


@pytest.fixture(params=["a", [], {}, None, 1])
def float_expected(request):
    """Return everything else"""
    return request.param


@pytest.fixture(params=[[], {}, None, [None], [{}], [[]]])
def parameter_value_expected(request):
    """Return everything else"""
    return request.param


@pytest.mark.parametrize("length", [0, 1, 2, 4, 5])
def test_component_is_too_short_or_too_long(length):
    """Test that a component has not enough values.

    See https://github.com/collective/icalendar/pull/979/#discussion_r2529703982

    """
    with pytest.raises(
        JCalParsingError, match="in Component: A component must be a list with 3 items."
    ):
        Component.from_jcal(["vevent", [], [], [], []][:length])


def test_invalid_component_type(list_expected):
    """We expect a list for components."""
    if isinstance(list_expected, str):
        return  # skip JSON
    with pytest.raises(
        JCalParsingError, match="Component: A component must be a list with 3 items."
    ):
        Component.from_jcal(list_expected)
    with pytest.raises(
        JCalParsingError,
        match="\\[2\\]\\[0\\] in Calendar: A component must be a list.",
    ):
        Component.from_jcal(["VCALENDAR", [], [list_expected]])


def test_invalid_component_name(str_expected):
    """Test the component name."""
    with pytest.raises(
        JCalParsingError, match="\\[0\\] in Component: The name must be a string."
    ):
        Component.from_jcal([str_expected, [], []])


def test_invalid_component_properties(list_expected):
    """Test the component properties."""
    with pytest.raises(
        JCalParsingError, match="\\[1\\] in Alarm: The properties must be a list."
    ):
        Component.from_jcal(["valarm", list_expected, []])


def test_invalid_component_subcomponents(list_expected):
    """Test the component subcomponents."""
    with pytest.raises(
        JCalParsingError, match="\\[2\\] in Todo: The subcomponents must be a list."
    ):
        Component.from_jcal(["vtodo", [], list_expected])


@pytest.mark.parametrize("i", [0, 3, 10])
def test_invalid_component_property(list_expected, i):
    """Test the component properties."""
    with pytest.raises(
        JCalParsingError,
        match=f"\\[1\\]\\[{i}\\] in Event: The property must be a list with at least 4 items.",
    ):
        Component.from_jcal(
            ["vevent", [["x-x", {}, "unknown", ""]] * i + [list_expected], []]
        )


@pytest.mark.parametrize("length", [0, 1, 2, 3])
def test_property_too_short(length, v_prop_example, v_prop):
    """The example is too short."""
    jcal = v_prop_example.to_jcal("name")[:length]
    with pytest.raises(
        JCalParsingError,
        match=f"in {v_prop.__name__}: The property must be a list with at least 4 items.",
    ):
        v_prop.from_jcal(jcal)

    with pytest.raises(
        JCalParsingError,
        match="in Parameters: The property must be a list with at least 4 items.",
    ):
        Parameters.from_jcal_property(jcal)


def test_property_name(v_prop_example, v_prop, str_expected):
    """The name is a string."""
    jcal = v_prop_example.to_jcal("name")
    jcal[0] = str_expected
    with pytest.raises(
        JCalParsingError,
        match=f"\\[0\\] in {v_prop.__name__}: The name must be a string.",
    ):
        v_prop.from_jcal(jcal)


def test_property_params(v_prop_example, v_prop, object_expected):
    """The name is a string."""
    jcal = v_prop_example.to_jcal("name")
    jcal[1] = object_expected
    with pytest.raises(
        JCalParsingError,
        match=f"\\[1\\] in {v_prop.__name__}: The parameters must be a mapping.",
    ):
        v_prop.from_jcal(jcal)
    with pytest.raises(
        JCalParsingError,
        match="\\[1\\] in Parameters: The parameters must be a mapping.",
    ):
        Parameters.from_jcal_property(jcal)


def test_property_type(v_prop_example, v_prop, str_expected):
    """The name is a string."""
    jcal = v_prop_example.to_jcal("name")
    jcal[2] = str_expected
    with pytest.raises(
        JCalParsingError,
        match=f"\\[2\\] in {v_prop.__name__}: The VALUE parameter must be a string.",
    ):
        v_prop.from_jcal(jcal)


@pytest.mark.parametrize("index", [0, 3])
def test_property_too_short_in_component(v_prop_example, v_prop, index):
    """The example is too short."""
    jcal = v_prop_example.to_jcal("X-PROP")
    component = [
        "vcalendar",
        [["X-PROP-2", {}, "unknown", ""]] * index + [jcal[:2]],
        [],
    ]
    print(jcal)
    with pytest.raises(
        JCalParsingError,
        match=f"\\[1\\]\\[{index}\\] in Calendar: The property must be a list with at least 4 items.",
    ):
        Component.from_jcal(component)


def test_parameters_keys(str_expected):
    """The parameter keys should be all strings."""
    if str_expected in ([], {}):
        return  # TypeError: unhashable type
    with pytest.raises(
        JCalParsingError,
        match="in Parameters: All parameter names must be strings.",
    ):
        Parameters.from_jcal_property(["", {str_expected: "value"}, "", ""])


@pytest.mark.parametrize("key", ["a", "key"])
def test_values_allowed_in_parameters(parameter_value_expected, key):
    """The parameter keys should be all strings."""
    with pytest.raises(
        JCalParsingError,
        match=f'\\[1\\]\\["{key}"\\] in Parameters: Parameter values must be a string, integer or float or a list of those.',
    ):
        Parameters.from_jcal_property(["", {key: parameter_value_expected}, "", ""])


def test_failing_example_delegated_to(calendars):
    """Check the failing example."""
    calendar: Calendar = calendars.rfc_7256_multi_value_parameters
    jcal = calendar.to_jcal()
    delegated_to = jcal[2][0][1][0][1]["delegated-to"]
    assert delegated_to == [
        "mailto:jdoe@example.com",
        "mailto:jqpublic@example.com",
    ]
    assert all(type(item) is str for item in delegated_to)


@pytest.mark.parametrize(
    ("v_prop", "ty", "message", "value"),
    [
        (vDatetime, "date-time", "Cannot parse date-time.", ""),
        (vDate, "date", "Cannot parse date.", ""),
        (vTime, "time", "Cannot parse time.", ""),
        (vTime, "time", "Cannot parse time.", "asd"),
        (vDuration, "duration", "Cannot parse duration.", "PXD"),
        (
            vDDDTypes,
            "date-time",
            "Cannot parse date, time, date-time, duration, or period.",
            "",
        ),
    ],
)
def test_date_time_parsing_errors(v_prop, message, ty, value):
    """An empty datetime should raise an error."""
    with pytest.raises(
        JCalParsingError,
        match=f"\\[3\\] in {v_prop.__name__}: {message}",
    ):
        v_prop.from_jcal(["dt", {}, ty, value])


@pytest.mark.parametrize(
    "v_prop",
    [
        vDuration,
        vDate,
        vTime,
        vDatetime,
        vDDDTypes,
    ],
)
def test_wrong_type(v_prop, str_expected):
    """Passing an int or float where a string is expected should raise an error."""
    if isinstance(str_expected, list) and v_prop is vDDDTypes:
        return  # skip vPeriod
    with pytest.raises(
        JCalParsingError,
        match=f"\\[3\\] in {v_prop.__name__}: The value must be a string.",
    ):
        v_prop.from_jcal(["dt", {}, "date-time", str_expected])


def test_vPeriod_wrong_type(str_expected):
    """Passing an int or float where a string is expected should raise an error."""
    with pytest.raises(
        JCalParsingError,
        match="\\[3\\]\\[0\\] in .*: The value must be a string.",
    ):
        vPeriod.from_jcal(
            ["rdate", {}, "period", [str_expected, "2024-01-01T00:00:00Z"]]
        )
    with pytest.raises(
        JCalParsingError,
        match="\\[3\\]\\[1\\] in .*: The value must be a string.",
    ):
        vPeriod.from_jcal(
            ["rdate", {}, "period", ["2024-01-01T00:00:00Z", str_expected]]
        )


def test_vPeriod_too_short():
    """A period with too few items should raise an error."""
    with pytest.raises(
        JCalParsingError,
        match="\\[3\\] in vPeriod: A period must be a list with exactly 2 items.",
    ):
        vPeriod.from_jcal(["rdate", {}, "period", ["2024-01-01T00:00:00Z"]])


def test_vPeriod_too_long():
    """A period with too many items should raise an error."""
    with pytest.raises(
        JCalParsingError,
        match="\\[3\\] in vPeriod: A period must be a list with exactly 2 items.",
    ):
        vPeriod.from_jcal(
            [
                "rdate",
                {},
                "period",
                ["2024-01-01T00:00:00Z", "2024-01-02T00:00:00Z", "extra"],
            ]
        )


def test_vPeriod_expects_date_time_as_start():
    """vPeriod expects date-time as start but we hand in a duration."""
    with pytest.raises(
        JCalParsingError,
        match="\\[3\\]\\[0\\] in .*: Cannot parse date-time.",
    ):
        vPeriod.from_jcal(
            [
                "rdate",
                {},
                "period",
                ["P1D", "2024-01-02T00:00:00Z"],
            ]
        )


def test_vPeriod_expects_date_time_or_duration_as_second_item():
    """vPeriod expects date-time or duration as second item but we hand in a date."""
    with pytest.raises(
        JCalParsingError,
        match="\\[3\\]\\[1\\] in .*: Cannot parse date-time or duration.",
    ):
        vPeriod.from_jcal(
            [
                "rdate",
                {},
                "period",
                ["2024-01-01T00:00:00", "2024-01-02"],
            ]
        )


@pytest.mark.parametrize("index", [0, 4])
def test_invalid_category_type(str_expected, index):
    """The name is a string."""
    with pytest.raises(
        JCalParsingError,
        match=f"\\[{index + 3}\\] in vCategory: The value must be a string.",
    ):
        vCategory.from_jcal(["categories", {}, "text"] + [""] * index + [str_expected])


def test_validation_of_list():
    """Check the list validation."""
    JCalParsingError.validate_list_type(["a", "b", "c"], str, "test", ["path"])
    with pytest.raises(
        JCalParsingError,
        match="\\[3\\]\\[1\\] in test: Each item in the list must be a string.",
    ):
        JCalParsingError.validate_list_type(["a", 1, "c"], str, "test", path=3)


def test_recurrence_rule_must_be_mapping(object_expected):
    """The recurrence rule must be a mapping."""
    with pytest.raises(
        JCalParsingError,
        match="\\[3\\] in vRecur: The recurrence rule must be a mapping with string keys.",
    ):
        vRecur.from_jcal(["rrule", {}, "recur", object_expected])


def test_recurrence_rule_must_be_mapping_with_str(str_expected):
    """The recurrence rule must be a mapping."""
    if isinstance(str_expected, (dict, list)):
        return  # skip unhashable type
    with pytest.raises(
        JCalParsingError,
        match="\\[3\\] in vRecur: The recurrence rule must be a mapping with string keys.",
    ):
        vRecur.from_jcal(["rrule", {}, "recur", {str_expected: 1}])


@pytest.mark.parametrize(
    "key",
    [
        "COUNT",
        "INTERVAL",
        "BYSECOND",
        "BYMINUTE",
        "BYHOUR",
        "BYWEEKNO",
        "BYMONTHDAY",
        "BYYEARDAY",
    ],
)
@pytest.mark.parametrize("as_list", [False, True])
def test_int_parameters(int_expected, key, as_list):
    """The integer parameters must be integers."""
    correct_value = 4
    if as_list:
        int_expected = [int_expected]
        correct_value = [4]
    # parse correct value
    recur = vRecur.from_jcal(["rrule", {}, "recur", {key: correct_value}])
    assert recur[key] == correct_value or recur[key] == [correct_value]
    if int_expected == []:
        return
    # parse bad value
    with pytest.raises(
        JCalParsingError,
        match=f'\\[3\\]\\["{key}"\\](?:\\[0\\])? in vInt: The value must be an integer.',
    ):
        vRecur.from_jcal(["rrule", {}, "recur", {key: int_expected}])


@pytest.mark.parametrize("valid", [1, "1", "1L"])
def test_parse_jcal_value_month(valid):
    """Test parsing of vMonth jCal values."""
    month = vMonth.parse_jcal_value(valid)
    assert month == 1
    assert month.leap == (valid == "1L")


@pytest.mark.parametrize("invalid", ["abc", "", 1.5, [], {}, "1LL"])
def test_parse_jcal_value_month_invalid(invalid):
    """Test parsing of vMonth jCal values."""
    with pytest.raises(
        JCalParsingError,
        match="in vMonth: The value must be a string or an integer.",
    ):
        vMonth.parse_jcal_value(invalid)


def test_skip():
    """The skip parameter works."""
    assert vSkip.parse_jcal_value("OMIT") is vSkip.OMIT
    with pytest.raises(
        JCalParsingError,
        match="in vSkip: The value must be a valid skip value.",
    ):
        vSkip.parse_jcal_value("INVALID")


def test_frequency():
    """The FREQ parameter must be valid."""
    # parse correct value
    recur = vRecur.from_jcal(["rrule", {}, "recur", {"FREQ": "DAILY"}])
    assert recur["FREQ"] == "DAILY"
    # parse bad value
    with pytest.raises(
        JCalParsingError,
        match='\\[3\\]\\["FREQ"\\] in vFrequency: The value must be a valid frequency.',
    ):
        vRecur.from_jcal(["rrule", {}, "recur", {"FREQ": "INVALID"}])
